#include "pch.hpp"
#include "scoreline.hpp"
#include "MappedFile.hpp"
#include "SwrAddrDef.h"

#include "MinimalMemory.hpp"

#define MINIMAL_USE_PROCESSHEAPSTRING
#include "MinimalPath.hpp"

#pragma pack(1)

struct SCORELINE_HEADER {
	DWORD Magic;
	DWORD Version;
};

struct TRACKRECORD {
	BYTE ID1P;
	BYTE ID2P;
	BYTE WIN;
};

struct TRACKRECORD_SPEC {
	BYTE ID2P;
	BYTE WIN;
};

struct TRACKRECORD_SPEC_CHUNK {
	DWORD trCount;
	DWORD trTop;
	TRACKRECORD_SPEC trackRecord[TRACKRECORD_MAX];
};

struct SCORELINE_V1 {
	SCORELINE_HEADER header;
	DWORD scoreLine[SWRCHAR_MAX][SWRCHAR_MAX][2];
	DWORD trCount;
	DWORD trTop;
	TRACKRECORD trackRecord[TRACKRECORD_MAX];
};

struct SCORELINE_V2 {
	SCORELINE_HEADER header;
	DWORD scoreLine[SWRCHAR_MAX][SWRCHAR_MAX][2];
	DWORD trCount;
	DWORD trTop;
	TRACKRECORD trackRecord[TRACKRECORD_MAX];
	TRACKRECORD_SPEC_CHUNK trackRecordSpec[SWRCHAR_MAX];
};

static DWORD s_scoreLineTR  [SWRCHAR_MAX][SWRCHAR_MAX][2];
static DWORD (*s_pScoreLine)[SWRCHAR_MAX][SWRCHAR_MAX][2];

static Minimal::ProcessHeapStringT<char> s_srPath;
static MappedFile s_slFile;
static SCORELINE_V2 *s_sl;

static DWORD s_readMode;

static void ScoreLine_RefleshScoreLineTR()
{
	ZeroMemory(s_scoreLineTR, sizeof s_scoreLineTR);
	DWORD range;
	switch(s_readMode) {
	case SCORELINE_READL30:  range = 30; break;
	case SCORELINE_READL50:  range = 50; break;
	case SCORELINE_READL100: range = 100;break;
	}
	DWORD count = range > s_sl->trCount? s_sl->trCount: range;
	DWORD bottom = s_sl->trTop - count;
	if(bottom < 0) bottom += TRACKRECORD_MAX;
	for(DWORD i = 0; i < count; ++i) {
		DWORD index = (bottom + i) % TRACKRECORD_MAX;
		TRACKRECORD &tr = s_sl->trackRecord[index];
		s_scoreLineTR[tr.ID1P][tr.ID2P][tr.WIN]++;
	}
}

static void ScoreLine_RefleshScoreLineTRS()
{
	ZeroMemory(s_scoreLineTR, sizeof s_scoreLineTR);
	DWORD range;
	switch(s_readMode) {
	case SCORELINE_READL30S:  range = 30; break;
	case SCORELINE_READL50S:  range = 50; break;
	case SCORELINE_READL100S: range = 100;break;
	}
	for(int j = 0; j < SWRCHAR_MAX; ++j) {
		TRACKRECORD_SPEC_CHUNK *chunk = &s_sl->trackRecordSpec[j];
		DWORD count = range > chunk->trCount? chunk->trCount: range;
		DWORD bottom = chunk->trTop - count;
		if(bottom < 0) bottom += TRACKRECORD_MAX;
		for(DWORD i = 0; i < count; ++i) {
			DWORD index = (bottom + i) % TRACKRECORD_MAX;
			TRACKRECORD_SPEC &tr = chunk->trackRecord[index];
			s_scoreLineTR[j][tr.ID2P][tr.WIN]++;
		}
	}
}

void ScoreLine_Clear()
{
	ZeroMemory(s_sl->scoreLine, sizeof s_sl->scoreLine);
	s_sl->trTop = s_sl->trCount = 0;
}

void ScoreLine_Append(int index1P, int index2P, int win)
{
	s_sl->scoreLine[index1P][index2P][win]++;

	s_sl->trackRecord[s_sl->trTop].ID1P = index1P;
	s_sl->trackRecord[s_sl->trTop].ID2P = index2P;
	s_sl->trackRecord[s_sl->trTop].WIN  = win;
	s_sl->trTop = (s_sl->trTop + 1) % TRACKRECORD_MAX;
	if(s_sl->trCount < TRACKRECORD_MAX) s_sl->trCount++;

	TRACKRECORD_SPEC_CHUNK *chunk = &s_sl->trackRecordSpec[index1P]; 
	chunk->trackRecord[chunk->trTop].ID2P = index2P;
	chunk->trackRecord[chunk->trTop].WIN  = win;
	chunk->trTop = (chunk->trTop + 1) % TRACKRECORD_MAX;
	if(chunk->trCount < TRACKRECORD_MAX)
		chunk->trCount++;

	switch(s_readMode) {
	case SCORELINE_READL30:
	case SCORELINE_READL50:
	case SCORELINE_READL100:
		ScoreLine_RefleshScoreLineTR();
		break;
	case SCORELINE_READL30S:
	case SCORELINE_READL50S:
	case SCORELINE_READL100S:
		ScoreLine_RefleshScoreLineTRS();
		break;
	}
}

int ScoreLine_Read(int index1P, int index2P, int win)
{
	return (*s_pScoreLine)[index1P][index2P][win];
}

DWORD& ScoreLine_Reference(int index1P, int index2P, int win)
{
	return (*s_pScoreLine)[index1P][index2P][win];
}
int ScoreLine_GetReadMode()
{
	return s_readMode;
}

void ScoreLine_SetReadMode(int mode)
{
	s_readMode = mode;
	switch(mode) {
	case SCORELINE_READALL:
		s_pScoreLine = &s_sl->scoreLine;
		break;
	case SCORELINE_READL30:
	case SCORELINE_READL50:
	case SCORELINE_READL100:
		s_pScoreLine = &s_scoreLineTR;
		ScoreLine_RefleshScoreLineTR();
		break;
	case SCORELINE_READL30S:
	case SCORELINE_READL50S:
	case SCORELINE_READL100S:
		s_pScoreLine = &s_scoreLineTR;
		ScoreLine_RefleshScoreLineTRS();
		break;
	}
}

void TrackRecord_Clear()
{
	::ZeroMemory(s_sl->trackRecord, sizeof s_sl->trackRecord);
	::ZeroMemory(s_scoreLineTR, sizeof s_scoreLineTR);
	s_sl->trCount = s_sl->trTop = 0;
	for(int i = 0; i < SWRCHAR_MAX; ++i) {
		s_sl->trackRecordSpec[i].trCount =
		s_sl->trackRecordSpec[i].trTop = 0;
	}
}

void ScoreLine_V1ToV2(MappedFile &slFile)
{
	SCORELINE_V1 sl;
	SCORELINE_V2 *nsl;

	slFile.Remap(sizeof SCORELINE_V1);
	void *p = slFile.Map(0, 0);
	memcpy(&sl, p, sizeof SCORELINE_V1);
	slFile.Unmap(p);

	slFile.Remap(sizeof SCORELINE_V2);
	nsl = (SCORELINE_V2 *)slFile.Map(0, 0);
	memcpy(nsl->scoreLine, sl.scoreLine, sizeof sl.scoreLine);
	memcpy(nsl->trackRecord, sl.trackRecord, sizeof sl.trackRecord);
	nsl->trCount = sl.trCount;
	nsl->trTop   = sl.trTop;

	memset(nsl->trackRecordSpec, 0, sizeof nsl->trackRecordSpec);
	int bottom = sl.trTop - sl.trCount;
	if(bottom < 0) bottom += TRACKRECORD_MAX;
	for(DWORD i = 0; i < sl.trCount; ++i) {
		int idx = (i + bottom) % TRACKRECORD_MAX;
		TRACKRECORD_SPEC_CHUNK *chunk =
			&nsl->trackRecordSpec[sl.trackRecord[idx].ID1P];
		chunk->trackRecord[chunk->trTop].ID2P = sl.trackRecord[idx].ID2P;
		chunk->trackRecord[chunk->trTop].WIN  = sl.trackRecord[idx].WIN;
		chunk->trCount = ++chunk->trTop;
	}
	nsl->header.Version = 2;
	slFile.Unmap(nsl);
}

bool ScoreLine_Map(LPCSTR fileName, bool create)
{
	MappedFile slFile;
	SCORELINE_HEADER *hed;
	SCORELINE_V2 *sl;

	if(!slFile.Open(fileName, 0, OPEN_EXISTING)) {
		if(!create || !slFile.Open(fileName, sizeof SCORELINE_V2)) return false;
		hed = (SCORELINE_HEADER *)slFile.Map(0, sizeof SCORELINE_V2);
		ZeroMemory(hed, sizeof SCORELINE_V2);
		hed->Magic = 'HISL';
		hed->Version = 2;
	} else {
		if(slFile.GetSize() < sizeof SCORELINE_HEADER) return false;
		hed = (SCORELINE_HEADER *)slFile.Map(0, sizeof SCORELINE_HEADER);
	}
	if(hed->Magic != 'HISL') {
		slFile.Unmap(hed);
		return false;
	}
	DWORD Version = hed->Version;
	slFile.Unmap(hed);
	switch(Version) {
	case 1: /* Ro[gΏ */
		ScoreLine_V1ToV2(slFile);
		break;
	case 2:	/* Ȃ */
		break;
	default:
		return false;
	}

	slFile.Remap(sizeof SCORELINE_V2);
	sl = (SCORELINE_V2 *)slFile.Map(0, 0);

	s_srPath = fileName;
	s_slFile.Unmap(s_sl);
	s_slFile = slFile;
	s_sl = sl;
/*
	long x = 1;
	for(int i = 0; i < SWRCHAR_MAX; ++i) {
		sl->trackRecordSpec[i].trCount = 100;
		for(int j = 0; j < 100; ++j) {
			x = x * 214013 + 2531011;
			int r = (x >> 16) & 32767;
			sl->trackRecordSpec[i].trackRecord[j].ID2P = (r >> 1) % SWRCHAR_MAX;
			sl->trackRecordSpec[i].trackRecord[j].WIN  = (r >> 6) % 2;
		}
	} */
	ScoreLine_SetReadMode(SCORELINE_READALL);
	return true;
}

LPCSTR ScoreLine_GetPath()
{
	return s_srPath;
}

void ScoreLine_Unmap()
{
	s_slFile.Unmap(s_sl);
	s_slFile.Close();
}
